@@ -26,15 +26,16 @@
 #include <stdint.h>
 #include <assert.h>
 #include <sys/param.h>
+#include "xvasprintf.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
-#include "xalloc.h"
+#include "xalloc.h"		/* Gnulib */
 #include "byteq.h"
-#include "error.h"
 #include "full-read.h"
 #include "full-write.h"
 
+#ifdef HAVE_FOPENCOOKIE
 static ssize_t byteq_libc_write(void *bq, const char *buf, size_t size);
 
 /* Function table for use with GNU libc's fopencookie.
@@ -54,6 +55,7 @@ byteq_libc_write(void *bq, const char *buf, size_t size)
     byteq_append(bq, buf, size);
     return size;
 }
+#endif
 
 /* Arbitrary enlargement of the byteq. This function only guarantees tha
  * the structure will hold one more byte after the call.
@@ -74,17 +76,13 @@ byteq_new(uint32_t initial_size)
     ByteQ *bq;
 
     bq = xmalloc(sizeof(ByteQ));
+#ifdef HAVE_FOPENCOOKIE
     bq->file = fopencookie(bq, "w", byteq_libc_io);
-    if (bq->file == NULL) {
-    	free(bq);
-	die(_("fopencookie failed - %s\n"), errstr);
-    }
-    if (setvbuf(bq->file, NULL, _IONBF, 0) != 0) {
-    	if (fclose(bq->file) != 0)
-	    warn(_("fclose failed - %s\n"), errstr);
-    	free(bq);
-	die(_("fopencookie failed - %s\n"), errstr);
-    }
+    if (bq->file == NULL)
+	xalloc_die();
+    if (setvbuf(bq->file, NULL, _IONBF, 0) != 0)
+	xalloc_die();
+#endif
     
     bq->cur = 0;
     bq->max = MAX(1, initial_size);
@@ -99,8 +97,9 @@ void
 byteq_free(ByteQ *bq)
 {
     if (bq != NULL) {
-	if (fclose(bq->file) < 0)
-    	    warn(_("fclose failed - %s\n"), errstr);
+#ifdef HAVE_FOPENCOOKIE
+	fclose(bq->file); /* ignore errors */
+#endif
 	free(bq->buf);
 	free(bq);
     }
@@ -134,7 +133,18 @@ byteq_append(ByteQ *bq, const uint8_t *data, uint32_t len)
 int
 byteq_vappendf(ByteQ *bq, const char *format, va_list ap)
 {
+#ifdef HAVE_FOPENCOOKIE
     return vfprintf(bq->file, format, ap);
+#else
+    char *str;
+    int len;
+
+    str = xvasprintf(format, ap);
+    len = strlen(str);
+    byteq_append(bq, str, len);
+    free(str);
+    return len;
+#endif
 }
 
 /* Append bytes to the byteq, printf style.
