@@ -52,7 +52,7 @@ void hmap_iterator(HMap *map, HMapIterator *it);
 void hmap_foreach_key(HMap *map, void (*iterator)());
 void hmap_foreach_value(HMap *map, void (*iterator)());
 void hmap_clear(HMap *map);
-uint32_t hmap_size(HMap *map);
+size_t hmap_size(HMap *map);
 void hmap_set_hash_fn(HMap *map, hash_fn_t hash);
 void hmap_set_compare_fn(HMap *map, comparison_fn_t compare);
 
