@@ -71,19 +71,22 @@ ptrv_append(PtrV *pv, void *value)
     pv->cur++;
 }
 
-/* Prepend a pointer to the ptrv.
+/* Prepend a few values to the ptrv.
  */
 void
-ptrv_prepend(PtrV *pv, void *value)
+ptrv_prepend_n(PtrV *pv, uint32_t count, void *value)
 {
-    if (pv->cur >= pv->max) {
-        pv->max = MAX(1, pv->max * 2);
-        pv->buf = xrealloc(pv->buf, pv->max * sizeof(void *));
-    }
-    memmove(pv->buf+1, pv->buf, pv->cur * sizeof(void *));
+    uint32_t c;
 
-    pv->buf[0] = value;
-    pv->cur++;
+    if (pv->cur+count > pv->max) {
+        pv->max = MAX(pv->cur+count, pv->max*2);
+        pv->buf = xrealloc(pv->buf, pv->max*sizeof(void *));
+    }
+    memmove(pv->buf+count, pv->buf, pv->cur*sizeof(void *));
+
+    for (c = 0; c < count; c++)
+	pv->buf[c] = value;
+    pv->cur += count;
 }
 
 /* Remove a range of pointers from the ptrv.
