@@ -43,10 +43,12 @@ void ptrv_foreach(PtrV *pv, PtrVForeachCallback callback);
 void ptrv_clear(PtrV *pv);
 
 /* The bad ones */
-void ptrv_remove_range(PtrV *pv, uint32_t start, uint32_t end);
+void ptrv_remove_range(PtrV *pv, uint32_t start, uint32_t end); /* XXX: rename ptrv_remove_n? */
 void *ptrv_remove_first(PtrV *pv);
 void ptrv_insort(PtrV *pv, void *element, comparison_fn_t comparator);
-void ptrv_prepend(PtrV *pv, void *value);
+#define ptrv_prepend(p,v) ptrv_prepend_n(p,1,v);
+void ptrv_prepend_n(PtrV *pv, uint32_t count, void *value);
+
 int32_t ptrv_find(PtrV *pv, void *element, comparison_fn_t comparator);
 
 #endif
