@@ -19,8 +19,6 @@
 
 #include "substrcmp.h"
 
-/* XXX: merge with strleftcmp? */
-
 int
 substrcmp(const char *s1, const char *s2, size_t s2len)
 {
@@ -31,5 +29,5 @@ substrcmp(const char *s1, const char *s2, size_t s2len)
 	s2++;
 	s2len--;
     }
-    return *s1;
+    return *s1; /* strncmp would return 0 here */
 }
