@@ -32,6 +32,7 @@
 #include "minmax.h"		/* Gnulib */
 #include "full-write.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
+#include "xstrndup.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
@@ -41,6 +42,7 @@
 #include "common/intutil.h"
 #include "common/strbuf.h"
 #include "common/strleftcmp.h"
+#include "common/substrcmp.h"
 #include "common/comparison.h"
 #include "microdc.h"
 
@@ -175,15 +177,14 @@ filelist_lookup(DCFileList *node, const char *filename)
     }
     if (node->type != DC_TYPE_DIR)
     	return NULL;
-    if (strncmp(filename, "..", end-filename) == 0) {
+    if (substrcmp("..", filename, end-filename) == 0) {
     	DCFileList *parent = (node->parent == NULL ? node : node->parent);
        	return (*end == '\0' ? parent : filelist_lookup(parent, end));
     }
-    if (strncmp(filename, ".", end-filename) == 0)
+    if (substrcmp(".", filename, end-filename) == 0)
     	return (*end == '\0' ? node : filelist_lookup(node, end));
 
-    /* FIXME. what's this? Why not do a real lookup instead?! */
-    name = strndup(filename, end-filename);
+    name = xstrndup(filename, end-filename); /* XXX: strndupa? */
     node = hmap_get(node->u.dir.children, name);
     free(name);
     if (node != NULL)
@@ -279,7 +280,7 @@ filelist_list_recursively(DCFileList *node, char *basepath, int skip)
 	char *tmp;
 
 	tmp = catfiles(basepath, node->name);
-	screen_putf("%7" PRIu64 "M %s%s\n", node->u.reg.size/(1024*1024), quotearg(tmp), "");
+	screen_putf("%7" PRIu64 "M %s%s\n", (uint64_t) (node->u.reg.size/(1024*1024)), quotearg(tmp), "");
 	free(tmp);
     }
 }
@@ -290,11 +291,11 @@ filelist_list(DCFileList *node, bool long_mode)
     uint32_t maxlen;
     uint64_t maxsize;
 
-    maxlen = 0;
-    maxsize = 0;
     if (node->type == DC_TYPE_DIR) {
         HMapIterator it;
 
+	maxlen = 0;
+	maxsize = 0;
 	hmap_iterator(node->u.dir.children, &it);
 	while (it.has_next(&it)) {
     	    DCFileList *subnode = it.next(&it);
@@ -318,7 +319,7 @@ filelist_list(DCFileList *node, bool long_mode)
     if (long_mode) {
         char *format;
 
-        format = xasprintf("%%%dlluM %%s%%s\n", maxsize == 0 ? 1 : ilog10(maxsize/(1024*1024)));
+        format = xasprintf("%%%d" PRIu64 "M %%s%%s\n", ilog10(max(1, maxsize/(1024*1024))));
 	if (node->type == DC_TYPE_DIR) {
 	    DCFileList **items;
 	    uint32_t c;
@@ -327,16 +328,16 @@ filelist_list(DCFileList *node, bool long_mode)
 	    for (c = 0; items[c] != NULL; c++) {
 		switch (items[c]->type) {
 		case DC_TYPE_REG:
-    		    screen_putf(format, items[c]->u.reg.size/(1024*1024), quotearg(items[c]->name), "");
+    		    screen_putf(format, (uint64_t) (items[c]->u.reg.size/(1024*1024)), quotearg(items[c]->name), "");
 		    break;
 		case DC_TYPE_DIR:
-    		    screen_putf(format, items[c]->u.dir.totalsize/(1024*1024), quotearg(items[c]->name), "/");
+    		    screen_putf(format, (uint64_t) (items[c]->u.dir.totalsize/(1024*1024)), quotearg(items[c]->name), "/");
 		    break;
 		}
 	    }
 	    free(items);
 	} else {
-	    screen_putf(format, node->u.reg.size, quotearg(node->name), "");
+	    screen_putf(format, (uint64_t) (node->u.reg.size/(1024*1024)), quotearg(node->name), "");
 	}
 	free(format);
     } else {
@@ -682,108 +683,118 @@ translate_local_to_remote(const char *localname)
     return strbuf_free_to_string(sb);
 }
 
-static bool
-local_fs_completion_generator(const char *dir_part, const char *file_part, int state, DCCompletionEntry *ce, bool dirsonly)
+void
+local_fs_completion_generator(const char *line,int ws,int we,char *word,PtrV *results,DCFSCompletionFlags flags)
 {
-    static DIR *dir;
+    DIR *dir;
+    const char *file_part;
+    char *dir_part;
 
-    if (state == 0)
-	dir = opendir(*dir_part ? dir_part : ".");
+    get_file_dir_part(word,&dir_part,&file_part);
     
+    dir = opendir(dir_part[0] != '\0' ? dir_part : ".");
     if (dir != NULL) {
     	struct dirent *de;
 
 	while ((de = readdir(dir)) != NULL) { /* Ignore errors */
-	    if (*(de->d_name) == '.' && *file_part != '.')
+	    if (!(flags & DC_CPL_DOT) && *(de->d_name) == '.' && *file_part != '.')
 		continue;
 	    if (strleftcmp(file_part, de->d_name) == 0) {
+	        DCCompletionEntry *entry;
 		struct stat st;
 		char *full;
+		char *display;
+		char *input;
 
-		full = catfiles(*dir_part ? dir_part : ".", de->d_name);
-		if (lstat(full, &st) < 0) {
+		full = catfiles(dir_part[0] != '\0' ? dir_part : ".", de->d_name);
+		if (stat(full, &st) < 0) {
 		    free(full);
 		    continue;
 		}
 
+                entry = new_completion_entry(NULL, NULL);
 		if (S_ISDIR(st.st_mode)) {
-		    /*ce->input_char = '/';*/
-		    /*ce->display_char = '/';*/
-		    ce->display_format = "%s/";
-		    ce->input_char = '/';
-		} else if (S_ISLNK(st.st_mode)) {
+		    if (flags & DC_CPL_DIR) {
+		        input = catfiles_with_trailing_slash(dir_part, de->d_name);
+		        display = xasprintf("%s/", de->d_name);
+		        entry->input_char = '\0';
+                        entry->display = display;
+                        entry->input = input;
+                        ptrv_append(results, entry);
+                    }
+		/*} else if (S_ISLNK(st.st_mode)) {
 		    if (stat(full, &st) >= 0 && S_ISDIR(st.st_mode)) {
-			ce->input_single_full_format = "%s/";
-		    	ce->input_char = '\0';
-			/*ce->input_append_single_full = '/';*/
+			entry->input_single_full_format = "%s/";
+		    	entry->input_char = '\0';
 		    } else {
-	    	    	if (dirsonly) {
+	    	    	if (!(flags & DC_CPL_REG)) {
 			    free(full);
 		    	    continue;
 			}
-			ce->input_char = ' ';
-		    }
-		    /*ce->display_char = '@';*/
-		    ce->display_format = "%s@";
-		} else {
-	    	    if (dirsonly) {
-			free(full);
-		    	continue;
-		    }
-		    ce->input_char = ' ';
-		    if (access(full, X_OK) >= 0) {
-			ce->display_format = "%s*";
-		    	/*ce->display_char = '*';*/
-		    /*} else {*/
-		    	/*ce->display_char = '\0';*/
+			entry->input_char = ' ';
 		    }
+		    input = catfiles(dir_part, de->d_name);
+		    display = xasprintf("%s@", de->d_name);
+                    entry->display = display;
+                    entry->input = input;
+                    ptrv_append(results, entry);*/
+		} else if (S_ISREG(st.st_mode)) {
+		    if ((flags & DC_CPL_REG) || (flags & DC_CPL_EXE)) {
+		        bool executable = (access(full, X_OK) == 0);
+		        if (executable || !(flags & DC_CPL_EXE)) {
+                            if (executable && !(flags & DC_CPL_EXE))
+		                display = xasprintf("%s*", de->d_name);
+                            else
+                                display = xstrdup(de->d_name);
+                            entry->input_char = ' ';
+                            input = catfiles(dir_part, de->d_name);
+                            entry->display = display;
+                            entry->input = input;
+                            ptrv_append(results, entry);
+                        }
+                    }
 		}
-		ce->str = xquotestr(de->d_name);
+
 		free(full);
-		return true;
 	    }
 	}
     	closedir(dir); /* Ignore errors */
     }
-
-    return false;
 }
 
-bool
-local_path_completion_generator(const char *path, const char *base, int state, DCCompletionEntry *ce)
+void
+local_path_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    return local_fs_completion_generator(path, base, state, ce, false);
+    local_fs_completion_generator(line,ws,we,word,results,DC_CPL_REG|DC_CPL_DIR);
 }
 
-bool
-local_dir_completion_generator(const char *path, const char *base, int state, DCCompletionEntry *ce)
+void
+local_dir_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    return local_fs_completion_generator(path, base, state, ce, true);
+    local_fs_completion_generator(line,ws,we,word,results,DC_CPL_DIR);
 }
 
-static bool
-remote_fs_completion_generator(const char *dir_path, const char *file_part, int state, DCCompletionEntry *ce, bool dirsonly)
+static void
+remote_fs_completion_generator(const char *line,int ws,int we,char *word,PtrV *results, bool dirsonly)
 {
-    static DCFileList *dir;
-    static HMapIterator dir_it;
-    static uint32_t dir_entry_count;
+    const char *file_part;
+    char *dir_part;
+    DCFileList *dir;
+    HMapIterator dir_it;
+    uint32_t dir_entry_count;
+    char *abspath;
 
     if (browse_list == NULL)
-    	return false;
+    	return;
 
-    if (state == 0) {
-    	char *abspath = apply_cwd(dir_path);
-
-	dir = filelist_lookup(browse_list, abspath);
-	free(abspath);
-    	if (dir == NULL || dir->type != DC_TYPE_DIR)
-	    return false;
-	hmap_iterator(dir->u.dir.children, &dir_it);
-	dir_entry_count = 0;
-    }
-
-    if (dir == NULL)
-    	return false;
+    get_file_dir_part(word,&dir_part,&file_part);
+    abspath = apply_cwd(dir_part);
+    dir = filelist_lookup(browse_list, abspath);
+    free(abspath);
+    if (dir == NULL || dir->type != DC_TYPE_DIR)
+        return;
+    hmap_iterator(dir->u.dir.children, &dir_it);
+    dir_entry_count = 0;
 
     for (;;) {
     	DCFileList *node;
@@ -812,31 +823,34 @@ remote_fs_completion_generator(const char *dir_path, const char *file_part, int
 	    continue;
 
 	if (strleftcmp(file_part, node->name) == 0) {
-    	    ce->str = xquotestr(node->name);
+	    DCCompletionEntry *entry;
+
 	    if (node->type == DC_TYPE_DIR) {
-		ce->input_char = '/';
-		/*ce->display_char = '/';*/
+	        entry = new_completion_entry(NULL, NULL);
+		entry->display = xasprintf("%s/", node->name);
+		entry->input = catfiles_with_trailing_slash(dir_part, node->name);
+		entry->input_char = '\0';
 	    } else {
-		ce->input_char = ' ';
-		/*ce->display_char = '\0';*/
+	        entry = new_completion_entry(NULL, NULL);
+	        entry->display = xstrdup(node->name);
+		entry->input = catfiles(dir_part, node->name);
+		entry->input_char = ' ';
 	    }
-	    return true;
+	    ptrv_append(results, entry);
 	}
     }
-
-    return false;
 }
 
-bool
-remote_path_completion_generator(const char *path, const char *base, int state, DCCompletionEntry *ce)
+void
+remote_path_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    return remote_fs_completion_generator(path, base, state, ce, false);
+    return remote_fs_completion_generator(line,ws,we,word,results,false);
 }
 
-bool
-remote_dir_completion_generator(const char *path, const char *base, int state, DCCompletionEntry *ce)
+void
+remote_dir_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    return remote_fs_completion_generator(path, base, state, ce, true);
+    return remote_fs_completion_generator(line,ws,we,word,results,true);
 }
 
 char *
