@@ -53,79 +53,64 @@ send_my_info(void)
 	  my_share_size);
 }
 
-bool
-say_user_completion_generator(const char *base, int state, DCCompletionEntry *ce)
+void
+say_user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static HMapIterator it;
-
-    if (state == 0)
-    	hmap_iterator(hub_users, &it);
+    HMapIterator it;
 
+    hmap_iterator(hub_users, &it);
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
-	if (strleftcmp(base, ui->nick) == 0) {
-	    ce->str = xquotestr(ui->nick);
-	    ce->input_single_format = "%s:";
-	    /*ce->input_append_single = ':';*/
-	    return true;
+
+	if (strleftcmp(word, ui->nick) == 0) {
+	    DCCompletionEntry *entry;
+	    entry = new_completion_entry(ui->nick, NULL);
+	    entry->input_single_format = "%s:";
+	    ptrv_append(results, entry);
 	}
     }
-
-    return false;
 }
 
-char *
-user_or_myself_completion_generator(const char *base, int state)
+void
+user_or_myself_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static HMapIterator it;
+    HMapIterator it;
 
-    if (state == 0) {
-    	hmap_iterator(hub_users, &it);
-	if (strleftcmp(base, my_nick) == 0)
-	    return xquotestr(my_nick);
-    }
+    hmap_iterator(hub_users, &it);
+    if (strleftcmp(word, my_nick) == 0)
+        ptrv_append(results, new_completion_entry(my_nick, NULL));
 
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
-	if (strleftcmp(base, ui->nick) == 0)
-	    return xquotestr(ui->nick);
+	if (strleftcmp(word, ui->nick) == 0)
+	    ptrv_append(results, new_completion_entry(ui->nick, NULL));
     }
-
-    return NULL;
 }
 
-char *
-user_completion_generator(const char *base, int state)
+void
+user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static HMapIterator it;
-
-    if (state == 0)
-    	hmap_iterator(hub_users, &it);
+    HMapIterator it;
 
+    hmap_iterator(hub_users, &it);
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
-	if (strleftcmp(base, ui->nick) == 0)
-	    return xquotestr(ui->nick);
+	if (strleftcmp(word, ui->nick) == 0)
+	    ptrv_append(results, new_completion_entry(ui->nick, NULL));
     }
-
-    return NULL;
 }
 
-char *
-user_with_queue_completion_generator(const char *base, int state)
+void
+user_with_queue_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static HMapIterator it;
-
-    if (state == 0)
-    	hmap_iterator(hub_users, &it);
+    HMapIterator it;
 
+    hmap_iterator(hub_users, &it);
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
-	if (ui->download_queue->cur > 0 && strleftcmp(base, ui->nick) == 0)
-	    return xquotestr(ui->nick);
+	if (ui->download_queue->cur > 0 && strleftcmp(word, ui->nick) == 0)
+	    ptrv_append(results, new_completion_entry(ui->nick, NULL));
     }
-
-    return NULL;
 }
 
 /* Create a new user info structure, representing a user on the hub.
@@ -208,7 +193,7 @@ hub_putf(const char *format, ...)
 	return false;
     }
 
-    dump_command("-->", hub_sendq->buf+oldcur, hub_sendq->cur-oldcur);
+    dump_command(_("-->"), hub_sendq->buf+oldcur, hub_sendq->cur-oldcur);
 
     res = byteq_write(hub_sendq, hub_socket);
     if (res == 0 || (res < 0 && errno != EAGAIN)) {
@@ -234,7 +219,7 @@ hub_connect(struct sockaddr_in *addr)
     }
 
     /* Set non-blocking I/O on socket. */
-    if (fd_set_nonblock_flag(hub_socket, true) < 0) {
+    if (!fd_set_nonblock_flag(hub_socket, true)) {
         warn(_("Cannot set socket flags - %s\n"), errstr);
 	hub_disconnect();
 	return;
@@ -402,8 +387,10 @@ hub_handle_command(char *buf, uint32_t len)
 	}
     	ui = hmap_get(hub_users, token);
 	if (ui == NULL) {
-	    warn(_("Invalid $MyINFO message: Unknown user %s, ignoring\n"), quotearg(token));
-	    return;
+	    /*warn(_("Invalid $MyINFO message: Unknown user %s, ignoring\n"), quotearg(token));*/
+            ui = user_info_new(token);
+            ui->info_quered = true;
+            hmap_put(hub_users, ui->nick, ui);
 	}
 
     	token = strsep(&buf, "$");
