@@ -27,6 +27,7 @@
 #include <sys/socket.h>		/* POSIX */
 #include <sys/stat.h>		/* ? */
 #include <sys/types.h>		/* ? */
+#include <sys/wait.h>		/* POSIX: waitpid */
 #include <unistd.h>		/* POSIX */
 #include <inttypes.h>		/* ? */
 #include <fcntl.h>		/* ? */
@@ -107,6 +108,7 @@ fd_set write_fds;
 PtrV *delete_files = NULL;
 PtrV *delete_dirs = NULL;
 static PtrV *search_udpmsg_out;	/* pending outgoing search results */
+pid_t shell_child = -1;
 
 const char version_etc_copyright[] = "Copyright (C) 2004-2005 Oskar Liljeblad";
 
@@ -266,10 +268,11 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
     switch (id) {
     case DC_MSG_SCREEN_PUT: {
     	char *msg;
+    	uint32_t flag;
 
-    	ipc_get(ipc, id, IPC_STR, &msg, IPC_END);
+    	ipc_get(ipc, id, IPC_INT32, &flag, IPC_STR, &msg, IPC_END);
 	/* Don't quote what the user prints */
-    	screen_putf(_("User %s: %s"), quotearg(uc->name), msg);
+    	flag_putf(flag, _("User %s: %s"), quotearg(uc->name), msg);
 	free(msg);
     	return true;
     }
@@ -330,6 +333,9 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
 	}
         return ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
     }
+    case DC_MSG_GET_MY_NICK: {
+        return ipc_get(ipc, id, IPC_END) && ipc_put(ipc, id, IPC_STR, my_nick, IPC_END);
+    };
     case DC_MSG_RESOLVE_FILE: {
     	char *inname;
 	char *inbase;
@@ -464,21 +470,17 @@ get_user_conn_status(DCUserConn *uc)
     return true;
 }*/
 
-char *
-transfer_completion_generator(const char *base, int state)
+void
+transfer_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static HMapIterator it;
-
-    if (state == 0)
-    	hmap_iterator(user_conns, &it);
+    HMapIterator it;
 
+    hmap_iterator(user_conns, &it);
     while (it.has_next(&it)) {
     	DCUserConn *uc = it.next(&it);
-	if (strleftcmp(base, uc->name) == 0)
-	    return xquotestr(uc->name);
+	if (strleftcmp(word, uc->name) == 0)
+	    ptrv_append(results, new_completion_entry(uc->name, NULL));
     }
-
-    return NULL;
 }
 
 /* This is the actual signal handler, registered in main_init with
@@ -491,11 +493,13 @@ signal_received(int signal)
 
     signal_char = signal;
     if (write(signal_pipe[1], &signal, sizeof(uint8_t)) < sizeof(uint8_t)) {
-	/* Die only if the signal is fatal. */
-	if (signal == SIGTERM || signal == SIGINT)
+	/* Die only if the signal is fatal.
+         * If SIGCHLD is not delivered, we would be stuck in a
+         * state where the user cannot enter any commands.
+         */
+	if (signal == SIGTERM || signal == SIGINT || signal == SIGCHLD)
 	    die(_("Cannot write to signal pipe - %s\n"), errstr); /* die OK */
-	else
-	    warn(_("Cannot write to signal pipe - %s\n"), errstr);
+        warn(_("Cannot write to signal pipe - %s\n"), errstr);
     }
 }
 
@@ -521,6 +525,23 @@ read_signal_input(void)
     else if (signal == SIGINT) {
         screen_erase_and_new_line();
     }
+    else if (signal == SIGCHLD) {
+        pid_t child;
+        int status;
+
+        while ((child = waitpid(-1, &status, WNOHANG)) > 0) {
+            if (child == shell_child) {
+                screen_wakeup(WIFSIGNALED(status) && WTERMSIG(status) == SIGINT);
+                shell_child = -1;
+            } else {
+                /* Assume it was a child process */
+                if (!WIFEXITED(status))
+                    warn(_("User process exited with status %d\n"), status);
+            }
+        }
+        if (child < 0 && errno != ECHILD)
+            warn(_("Cannot wait for processes - %s\n"), errstr);
+    }
     else if (signal == SIGUSR1) {
         /* Not implemented yet, do nothing for now. */
     }
@@ -620,7 +641,7 @@ search_input_available(void)
 	return;
     }
 
-    dump_command("<==", search_recvq->buf, search_recvq->cur);
+    dump_command(_("<=="), search_recvq->buf, search_recvq->cur);
     search_recvq->buf[search_recvq->cur-1] = '\0'; /* not strictly necessary */
     handle_search_result(search_recvq->buf, search_recvq->cur);
     byteq_clear(search_recvq);
@@ -674,7 +695,7 @@ enable_search(void)
 	return false;
     }
 
-    if (fd_set_nonblock_flag(search_socket, true) < 0) {
+    if (!fd_set_nonblock_flag(search_socket, true)) {
     	warn(_("Cannot set socket flags - %s\n"), errstr);
 	return false;
     }
@@ -715,7 +736,7 @@ enable_active(uint16_t port)
     }
 
     /* Set non-blocking */
-    if (fd_set_nonblock_flag(listen_socket, true) < 0) {
+    if (!fd_set_nonblock_flag(listen_socket, true)) {
     	warn(_("Cannot set socket flags - %s\n"), errstr);
 	disable_active();
 	return false;
@@ -763,6 +784,7 @@ bool
 set_active(bool newactive, uint16_t port)
 {
     if (newactive) {
+	disable_active();
     	if (!enable_active(port))
 	    return false;
     } else {
@@ -834,6 +856,8 @@ main (int argc, char **argv)
 	case VERSION_OPT:
 	    version_etc(stdout, NULL, PACKAGE, VERSION, "Oskar Liljeblad", NULL);
 	    exit(EXIT_SUCCESS);
+        default:
+            exit(EXIT_FAILURE);
 	}
     }
     
@@ -861,6 +885,10 @@ main (int argc, char **argv)
         warn(_("Cannot register signal handler - %s\n"), errstr);
 	goto cleanup;
     }
+    if (sigaction(SIGCHLD, &sigact, NULL) < 0) {
+        warn(_("Cannot register signal handler - %s\n"), errstr);
+	goto cleanup;
+    }
     sigact.sa_handler = SIG_IGN;
     if (sigaction(SIGPIPE, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
@@ -870,7 +898,7 @@ main (int argc, char **argv)
     FD_ZERO(&read_fds);
     FD_ZERO(&write_fds);
     FD_SET(signal_pipe[0], &read_fds);
-    FD_SET(STDIN_FILENO, &read_fds);
+    /*FD_SET(STDIN_FILENO, &read_fds);*/
 
     hub_recvq = byteq_new(128);
     hub_sendq = byteq_new(128);
@@ -909,6 +937,7 @@ main (int argc, char **argv)
     	goto cleanup;
     my_ul_slots = 3;
 
+    command_init();
     get_package_file("config", &filename);
     run_script(filename, true);
     free(filename);
@@ -918,7 +947,7 @@ main (int argc, char **argv)
 	fd_set res_write_fds;
     	int res;
 
-    	screen_prepare();
+        screen_prepare();
 	res_read_fds = read_fds;
 	res_write_fds = write_fds;
     	res = TEMP_FAILURE_RETRY(select(FD_SETSIZE, &res_read_fds, &res_write_fds, NULL, NULL));
@@ -955,8 +984,8 @@ main (int argc, char **argv)
 cleanup:
 
     hub_disconnect();
-
     screen_finish();
+    command_finish();
 
     byteq_free(hub_recvq);
     byteq_free(hub_sendq);
