@@ -25,6 +25,7 @@
 #include <stdint.h>
 #include <netinet/in.h>
 #include <time.h>
+#include <fcntl.h>		/* POSIX: O_NONBLOCK */
 #include "common/byteq.h"
 #include "common/ptrv.h"
 #include "common/error.h"
@@ -38,6 +39,13 @@
 
 #define SEARCH_TIME_THRESHOLD 60        /* Add no more results to searches after this many seconds elapsed */
 
+typedef enum {
+    DC_CPL_DIR	= 1 << 0,	/* complete directories or symlinks to directories */
+    DC_CPL_REG	= 1 << 1,	/* complete regular files or symlinks to regular files */
+    DC_CPL_EXE	= 1 << 2,	/* complete regular files or symlinks to regular files that are executable according to access */
+    DC_CPL_DOT	= 1 << 3,	/* always complete files and dirs starting with `.', even if file part of `word' does not start with `.' */
+} DCFSCompletionFlags;
+
 typedef enum {
     DC_DF_DEBUG			= 1 << 0, /* Various debug messages */
     DC_DF_JOIN_PART		= 1 << 1, /* Joins and Quits messages */
@@ -46,16 +54,9 @@ typedef enum {
     DC_DF_UPLOAD		= 1 << 4, /* "Upload complete" message (not errors) */
     DC_DF_DOWNLOAD		= 1 << 5, /* "Download complete" message (not errors) */
     DC_DF_CONNECTIONS		= 1 << 6, /* User connections and normal disconnections */
+    DC_DF_COMMON		= 1 << 7, /* Common messages always displayed (XXX: get rid of this?) */
 } DCDisplayFlag;
 
-typedef enum {
-    DC_CPL_NONE,
-    DC_CPL_SIMPLE,
-    DC_CPL_FILE,
-    DC_CPL_CUSTOM,
-    DC_CPL_CHAIN,
-} DCCompletionFunctionType;
-
 typedef enum {
     DC_TF_NORMAL,		/* Normal file transfer */
     DC_TF_LIST,			/* Transfer of MyList.DcLst */
@@ -108,6 +109,7 @@ typedef enum {
     DC_MSG_WANT_DOWNLOAD,
     DC_MSG_VALIDATE_DIR,
     DC_MSG_VALIDATE_NICK,
+    DC_MSG_GET_MY_NICK,
     DC_MSG_RESOLVE_FILE, /* for download or upload */
     DC_MSG_STATUS,
     DC_MSG_DISCONNECT,
@@ -139,6 +141,10 @@ typedef struct _DCSearchResponse DCSearchResponse;
 typedef struct _DCQueuedFile DCQueuedFile;
 typedef struct _DCVariable DCVariable;
 
+typedef void (*DCCompletorFunction)(const char *line,int ws,int we,char *word,PtrV *results);
+typedef void (*DCBuiltinCommandHandler)(int argc, char **argv);
+typedef void (*ScreenWriter)(DCDisplayFlag flag, const char *format, va_list args);
+
 struct _DCSearchResponse {
     uint32_t refcount;
     DCUserInfo *userinfo;
@@ -178,12 +184,12 @@ struct _DCSearchRequest {
 };
 
 struct _DCCompletionEntry {
-    char *str;
-    char input_char;
+    char *input;
     char *input_format;
-    char *display_format;
+    char *display;
     char *input_single_format;
     char *input_single_full_format;
+    char input_char;
 //  char input_char; /* character to append to input line */
 //  char display_char; /* character to display */
 //  char input_append_single; /* append to line buffer when this entry is the only match */
@@ -252,25 +258,15 @@ struct _DCUserConn {
 struct _DCVariable {
     char *name;
     char *(*getter)(DCVariable *var);
-    void (*setter)(DCVariable *var, const char *new_value);
+    void (*setter)(DCVariable *var, int argc, char **argv); /* argc >= 1 */
     void *value;
-    DCCompletionFunctionType completor_type;
-    void *completor;
+    DCCompletorFunction completor;
     void *type_details;
     char *help_string;
 };
 
-/* We provide two completion function prototypes:
- * One for more complex completors (file completors),
- * and one for less complex completors.
- */
-typedef char *(*DCCompletionFunction)(const char *base, int state);
-typedef bool (*DCFileCompletionFunction)(const char *full,
-	const char *base, int state, DCCompletionEntry *ce);
-typedef bool (*DCCustomCompletionFunction)(const char *base, int state, DCCompletionEntry *ce);
-typedef void *(*DCChainCompletionFunction)(const char *buffer, int start, int end, DCCompletionFunctionType *type);
 
-extern vprintf_fn_t screen_writer;
+extern ScreenWriter screen_writer;
 extern char *log_filename;
 
 extern DCHubState hub_state;
@@ -312,6 +308,8 @@ extern bool browsing_myself;
 extern char *browse_path;
 extern char *browse_path_previous; /* recorded for `cd -' purposes */
 
+extern pid_t shell_child;
+
 /* hub.c */
 void hub_input_available(void);
 bool hub_putf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));
@@ -325,30 +323,37 @@ void hub_now_writable(void);
 bool send_my_info(void);
 extern struct sockaddr_in hub_addr;
 extern char *hub_name;
-bool say_user_completion_generator(const char *base, int state, DCCompletionEntry *ce);
-char *user_completion_generator(const char *base, int state);
-char *user_or_myself_completion_generator(const char *base, int state);
-char *user_with_queue_completion_generator(const char *base, int state);
+void say_user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void user_or_myself_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void user_with_queue_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
 
 /* command.c */
-void command_execute(char *line);
-void *default_completion_selector(const char *buffer, int start, int end, DCCompletionFunctionType *type);
+void command_execute(const char *line);
+void default_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
 /*bool append_download_dir(DCUserInfo *ui, DCFileList *node, char *base_path, uint64_t *bytes, uint32_t *files) XXX move transfer.c? */
 void update_prompt(void);
+void command_init(void);
+void command_finish(void);
 
 /* screen.c */
-void screen_putf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-void screen_vputf(const char *format, va_list args) __attribute__ ((format (printf, 1, 0)));
-void flag_putf(DCDisplayFlag flag, const char *format, ...);
+#define screen_putf(f,...) flag_putf(DC_DF_COMMON, (f), ## __VA_ARGS__)
+void flag_putf(DCDisplayFlag flag, const char *format, ...) __attribute__ ((format (printf, 2, 3)));
 void screen_erase_and_new_line(void);
 void screen_finish(void);
+void screen_suspend(void);
+void screen_wakeup(bool print_newline_first);
 void screen_prepare(void);
 void screen_read_input(void);
 void screen_get_size(int *rows, int *cols);
 void set_screen_prompt(const char *prompt, ...) __attribute__ ((format (printf, 1, 2)));
 bool set_log_file(const char *new_filename, bool verbose);
-char *sorted_list_completion_generator(const char *base, int state,
-    void *items, size_t item_count, size_t item_size, size_t key_offset);
+void sorted_list_completion_generator(const char *base, PtrV *results,
+    void *items, size_t item_count, size_t item_size, size_t key_offset); /* completion */
+DCCompletionEntry *new_completion_entry(const char *input, const char *display); /* completion */
+void free_completion_entry(DCCompletionEntry *entry); /* completion */
+void get_file_dir_part(const char *word,char **dir_part,const char **file_part); /* completion */
+
 
 /* huffman.c */
 uint8_t *huffman_decode(const uint8_t *data, uint32_t data_size, uint32_t *out_size);
@@ -367,7 +372,7 @@ bool set_share_dir(const char *dir);
 bool has_user_conn(DCUserInfo *info, DCTransferDirection dir);
 uint32_t get_user_conn_count(DCUserInfo *info);
 bool get_package_file(const char *name, char **outname);
-char *transfer_completion_generator(const char *base, int state); /* XXX: move transfer.c? */
+void transfer_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
 void user_conn_cancel(DCUserConn *uc);
 void warn_file_error(int res, bool write, const char *filename);
 void warn_socket_error(int res, bool write, const char *subject);
@@ -385,10 +390,11 @@ char *resolve_download_file(DCUserInfo *ui, const char *inname, const char *base
 bool filelist_create(const char *dir);
 char *translate_local_to_remote(const char *localname);
 char *translate_remote_to_local(const char *remotename);
-bool local_path_completion_generator(const char *full, const char *base, int state, DCCompletionEntry *ce);
-bool local_dir_completion_generator(const char *full, const char *base, int state, DCCompletionEntry *ce);
-bool remote_path_completion_generator(const char *full, const char *base, int state, DCCompletionEntry *ce);
-bool remote_dir_completion_generator(const char *full, const char *base, int state, DCCompletionEntry *ce);
+void local_fs_completion_generator(const char *line,int ws,int we,char *word,PtrV *results, DCFSCompletionFlags flags);
+void local_path_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void local_dir_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void remote_path_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void remote_dir_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
 char *apply_cwd(const char *path); /* XXX: move transfer.c? */
 extern DCFileList *our_filelist;
 void filelist_list_recursively(DCFileList *node, char *basepath, int skip);
@@ -402,16 +408,23 @@ void dump_command(const char *header, const char *buf, int len);
 int ilog10(uint64_t c);
 int mkdirs_for_file(char *filename, bool deletedirs);
 char *catfiles(const char *p1, const char *p2);
-int fd_set_nonblock_flag(int fd, int value);
+char *catfiles_with_trailing_slash(const char *p1, const char *p2);
+bool fd_set_status_flags(int fd, bool set, int modflags);
+#define fd_set_nonblock_flag(f,s) fd_set_status_flags(f,s,O_NONBLOCK)
 char *getenv_default(const char *name, char *defvalue);
 #define IS_CURRENT_DIR(x) ((x)[0] == '.' && (x)[1] == '\0')
 #define IS_PARENT_DIR(x)  ((x)[0] == '.' && (x)[1] == '.' && (x)[2] == '\0')
 #define COMPARE_RETURN(a,b) { if ((a) < (b)) return -1; if ((a) > (b)) return 1; }
 #define COMPARE_RETURN_FUNC(f) { int _c = (f); if (_c != 0) return _c; }
 char *sockaddr_in_str(struct sockaddr_in *addr);
+char *in_addr_str(struct in_addr addr);
 bool parse_ip_and_port(char *source, struct sockaddr_in *addr, uint16_t defport);
 #define xquotestr(s) 		(quotearg_alloc((s), SIZE_MAX, NULL))
 #define quotearg_mem(m,l)	(quotearg_n_style_mem(0, escape_quoting_style, (m), (l)))
+#if 0
+char *xconfstr(const char *name);
+#endif
+char *join_strings(char **strs, int count, char mid);
 
 /* search.c */
 bool parse_search_selection(char *str, DCSearchSelection *data);
@@ -423,7 +436,7 @@ void free_search_request(DCSearchRequest *sr);
 char *search_selection_to_string(DCSearchSelection *sr);
 
 /* variables.c */
-void cmd_set(char *cmd, char *args);
-void *set_completion_selector(const char *buffer, int start, int end, DCCompletionFunctionType *type);
+void cmd_set(int argc, char **argv);
+void set_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
 
 #endif
