@@ -18,6 +18,7 @@
  */
 
 #include <config.h>
+#include <unistd.h>		/* POSIX: STDIN_FILENO */
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
@@ -45,244 +46,44 @@
 #include "common/bksearch.h"
 #include "common/quoting.h"
 
+/* Flow of messages in microdc:
+ *
+ *  main process:
+ *    warn(error.c) -> warn_writer(error.c)  -> screen_warn_writer(screen.c) -> screen_writer(screen.c) -> flag_vputf(screen.c) -> log/screen
+ *    die(error.c)  -> warn_writer(error.c)  -> screen_warn_writer(screen.c) -> screen_writer(screen.c) -> flag_vputf(screen.c) -> log/screen
+ *                     screen_putf(screen.c) -> flag_putf(screen.c)          -> screen_writer(screen.c) -> flag_vputf(screen.c) -> log/screen
+ *  user process:
+ *    warn(error.c) -> warn_writer(error.c)  -> screen_warn_writer(screen.c) -> screen_writer(screen.c) -> user_screen_writer(user.c) -> ipc
+ *    die(error.c)  -> warn_writer(error.c)  -> screen_warn_writer(screen.c) -> screen_writer(screen.c) -> user_screen_writer(user.c) -> ipc
+ *                     screen_putf(screen.c) -> flag_putf(screen.c)          -> screen_writer(screen.c) -> user_screen_writer(user.c) -> ipc
+ *
+ */
+
+typedef enum {
+    SCREEN_UNINITIALIZED,	/* History not initialized */
+    SCREEN_NO_HANDLER,		/* rl_callback_handler_install not called. */
+    SCREEN_RL_CLEARED,		/* Readline has been cleared. Need to redisplay after. */
+    SCREEN_RL_DISPLAYED,	/* Readline will be displayed. Need to clear first. */
+    SCREEN_SUSPENDED,		/* A command using the screen is running. */
+} ScreenState;
+
 static void clear_rl();
 static void user_input(char *line);
-static int real_screen_vputf(const char *format, va_list args);
+static int screen_warn_writer(const char *format, va_list args);
+static void flag_vputf(DCDisplayFlag flag, const char *format, va_list args);
+
+char *log_filename = NULL;
+ScreenWriter screen_writer = flag_vputf;
 static char *screen_prompt;
 static FILE *log_fh = NULL;
-char *log_filename = NULL;
-
-vprintf_fn_t screen_writer = real_screen_vputf;
-
-enum {
-    SCREEN_NO_HISTORY,
-    SCREEN_NO_HANDLER,		/* rl_callback_handler_install not called. */
-    SCREEN_NORMAL,		/* Normal non-readline status. Can print right away. */
-    SCREEN_RL_DISPLAYED,	/* Readline will be displayed. Need to clear first. */
-    SCREEN_RL_CLEARED,		/* Readline has been cleared. Need to redisplay after. */
-} screen_state = SCREEN_NO_HISTORY;
-
-bool
-set_log_file(const char *new_filename, bool verbose)
-{
-    if (log_fh != NULL) {
-	if (fclose(log_fh) != 0)
-	    warn(_("%s: Cannot close file - %s\n"), quotearg(log_filename), errstr);
-	log_fh = NULL;
-	free(log_filename);
-	log_filename = NULL;
-    }
-    if (new_filename == NULL) {
-	if (verbose)
-	    screen_putf(_("No longer logging to file.\n"));
-	return true;
-    }
-    log_fh = fopen(new_filename, "a");
-    if (log_fh == NULL) {
-	screen_putf(_("%s: Cannot open file for appending - %s\n"), quotearg(new_filename), errstr);
-	return false;
-    }
-    log_filename = xstrdup(new_filename);
-    if (verbose)
-	screen_putf(_("Logging to `%s'.\n"), quotearg(new_filename));
-    return true;
-}
-
-/* Readline < 5.0 disables SA_RESTART on SIGWINCH for some reason.
- * This turns it back on.
- * This was copied from guile.
- */
-static int
-fix_winch(void)
-{
-    struct sigaction action;
-
-    if (sigaction(SIGWINCH, NULL, &action) >= 0) {
-    	action.sa_flags |= SA_RESTART;
-    	sigaction(SIGWINCH, &action, NULL); /* Ignore errors */
-    }
-
-    return 0;
-}
-
-static char **
-attempted_completion(const char *text, int start, int end)
-{
-    void *completor;
-    DCCompletionFunctionType type;
-    char **matches = NULL;
-    char *dir_part = NULL;
-    char *file_part = NULL;
-    char *text_nq;
-    bool quoted;
-
-    /* Determine if current word is quoted, and dequote. */
-    quoted = char_is_quoted(rl_line_buffer, start)
-              && rl_line_buffer[start-1] == '"';
-    text_nq = dequote_words(text, quoted, NULL);
-   
-    /* Do not attempt readline's default filename completion if our
-     * completors fails to return any results.
-     */
-    rl_attempted_completion_over = 1;
-
-    completor = default_completion_selector(rl_line_buffer, start, end, &type);
-    /* XXX: Better idea than DC_CPL_CHAIN: 
-     * Let the completor itself determine how to handle
-     * arguments. Is the dir/file_part separation really necessary?
-     * Cannot the completor itself do this itself? All that is
-     * necessary is rules before, and rules after. Shouldn't the
-     * completer return a complete set of results instead? And be
-     * passed a LList or something where results can be added?
-     * The first will always be the input result... But gcd (leading
-     * character) still need to be calculated.
-     */
-
-    if (type == DC_CPL_CHAIN) {
-	DCChainCompletionFunction chain_completor = completor;
-	completor = chain_completor(rl_line_buffer, start, end, &type);
-    }
-    if (type == DC_CPL_FILE) {
-	file_part = strrchr(text_nq, '/');
-	if (file_part == NULL) {
-	    dir_part = xstrdup("");
-	    file_part = text_nq;
-	} else {
-	    for (; file_part > text_nq && file_part[-1] == '/'; file_part--);
-	    dir_part = xstrndup(text_nq, file_part-text_nq+1);
-	    for (file_part++; *file_part == '/'; file_part++);
-	}
-    } else if (type == DC_CPL_SIMPLE || type == DC_CPL_CUSTOM) {
-	dir_part = NULL;
-	file_part = text_nq;
-    }
-
-    if (type == DC_CPL_NONE || completor != NULL) {
-    	PtrV *completion_entries;
-    	int state;
-	int lcd = INT_MAX;
-	uint32_t c;
-
-    	completion_entries = ptrv_new();
-	for (state = 0; true; state++) {
-	    /*DCCompletionEntry ce = { NULL, ' ', '\0', '\0', '\0' };*/
-	    DCCompletionEntry ce = { NULL, ' ', "%s", "%s", NULL, NULL };
-
-	    if (type == DC_CPL_SIMPLE) {
-		DCCompletionFunction simple_completor = completor;
-		ce.str = simple_completor(file_part, state);
-		if (ce.str == NULL)
-		    break;
-	    } else if (type == DC_CPL_FILE) {
-		DCFileCompletionFunction file_completor = completor;
-		if (!file_completor(dir_part, file_part, state, &ce))
-		    break;
-	    } else if (type == DC_CPL_CUSTOM) {
-		DCCustomCompletionFunction custom_completor = completor;
-		if (!custom_completor(file_part, state, &ce))
-		    break;
-	    }
-	    ptrv_append(completion_entries, xmemdup(&ce, sizeof(ce)));
-
-    	    if (state > 0) {
-    		DCCompletionEntry *oldce = completion_entries->buf[state-1];
-		int c1, c2, si;
-
-                /* Determine common leading characters for all matches. */
-	    	for (si = 0; (c1 = oldce->str[si]) && (c2 = ce.str[si]); si++) {
-		    if (c1 != c2)
-		    	break;
-    	    	}
-	    	lcd = MIN(lcd, si);
-	    }
-	}
-
-    	if (state == 0) { /* no matches */
-	    ptrv_free(completion_entries);
-	    matches = NULL;
-	}
-	else if (state == 1) { /* single match */
-	    char *tmp;
-	    DCCompletionEntry *ce;
-
-            ce = completion_entries->buf[0];
-	    matches = xmalloc(sizeof(char *) * 2);
-
-	    if (ce->input_single_format != NULL) {
-		tmp = xasprintf(ce->input_single_format, ce->str);
-		free(ce->str);
-		ce->str = tmp;
-	    } else if (ce->input_single_full_format != NULL && strcmp(file_part, ce->str) == 0) {
-		free(ce->str);
-		ce->str = xasprintf(ce->input_single_full_format, file_part);
-	    }
-
-	    if (type == DC_CPL_FILE) {
-		tmp = catfiles(dir_part, ce->str);
-		free(ce->str);
-		ce->str = xasprintf(ce->input_format, tmp);
-		free(tmp);
-	    } else if (type == DC_CPL_SIMPLE || type == DC_CPL_CUSTOM) {
-		tmp = xasprintf(ce->input_format, ce->str);
-		free(ce->str);
-		ce->str = tmp;
-	    }
-
-	    matches[0] = quote_word(ce->str, quoted, true);
-	    matches[1] = NULL;
-
-	    rl_completion_append_character = ce->input_char;
-	    free(ce->str);
-    	    free(ce);
-	    ptrv_free(completion_entries);
-	}
-    	else { /* multiple matches */
-    	    DCCompletionEntry *ce = completion_entries->buf[0];
-	    char *tmp = NULL;
-
-    	    matches = xmalloc(sizeof(char *) * (state + 2));
-	    matches[0] = xstrndup(ce->str, lcd);
-	    if (type == DC_CPL_FILE) {
-		char *tmp2;
-		tmp2 = catfiles(dir_part, matches[0]);
-		tmp = xasprintf(ce->input_format, tmp2);
-		free(tmp2);
-		free(matches[0]);
-	    }
-	    else if (type == DC_CPL_SIMPLE || type == DC_CPL_CUSTOM) {
-		tmp = xasprintf(ce->input_format, matches[0]);
-		free(matches[0]);
-	    }
-	    matches[0] = quote_word(tmp, quoted, false);
-	    free(tmp);
-
-	    for (c = 0; c < state; c++) {
-	    	ce = completion_entries->buf[c];
-		/*if (ce->display_format != NULL) {*/
-		    /*matches[c+1] = xasprintf("%s%c", ce->str, ce->display_char);*/
-		    matches[c+1] = xasprintf(ce->display_format, ce->str);
-		    free(ce->str);
-		/*} else {
-	    	    matches[c+1] = ce->str;
-		}*/
-	    }
-	    matches[state+1] = NULL;
-	    ptrv_foreach(completion_entries, free);
-	    ptrv_free(completion_entries);
-	}
-    }
-
-    free(text_nq);
-    free(dir_part);
-
-    return matches;
-}
+static PtrV *suspend_msgs = NULL;
+static ScreenState screen_state = SCREEN_UNINITIALIZED;
 
 /* This piece of code was snatched from lftp, and modified somewhat by me.
  * I suggest a function called rl_clear() be added to readline. The
  * function clears the prompt and everything the user has written so far on
  * the line. The cursor is positioned at the beginning of the line that
- * contained the prompt Note: This function doesn't modify the screen_state
+ * contained the prompt. Note: This function doesn't modify the screen_state
  * variable.
  */
 static void
@@ -314,29 +115,190 @@ clear_rl()
         rl_expand_prompt(rl_prompt);
 }
 
-static int
-real_screen_vputf(const char *format, va_list args)
+bool
+set_log_file(const char *new_filename, bool verbose)
 {
-    if (screen_state == SCREEN_RL_DISPLAYED) {
-        clear_rl();
-        screen_state = SCREEN_RL_CLEARED;
+    if (log_fh != NULL) {
+	if (fclose(log_fh) != 0)
+	    warn(_("%s: Cannot close file - %s\n"), quotearg(log_filename), errstr);
+	log_fh = NULL;
+	free(log_filename);
+	log_filename = NULL;
     }
-    return vprintf(format, args);
+    if (new_filename == NULL) {
+	if (verbose)
+	    screen_putf(_("No longer logging to file.\n"));
+	return true;
+    }
+    log_fh = fopen(new_filename, "a");
+    if (log_fh == NULL) {
+	screen_putf(_("%s: Cannot open file for appending - %s\n"), quotearg(new_filename), errstr);
+	return false;
+    }
+    log_filename = xstrdup(new_filename);
+    if (verbose)
+	screen_putf(_("Logging to `%s'.\n"), quotearg(new_filename));
+    return true;
 }
 
-/* Print something on screen, printf format style.
- * __attribute__ is provided by microdc.h declaration.
+/* Readline < 5.0 disables SA_RESTART on SIGWINCH for some reason.
+ * This turns it back on.
+ * This was partially copied from guile.
  */
-void
-screen_putf(const char *format, ...)
+static int
+fix_winch(void)
 {
-    va_list args;
+    struct sigaction action;
 
-    va_start(args, format);
-    screen_writer(format, args);
-    if (log_fh != NULL)
-	vfprintf(log_fh, format, args);
-    va_end(args);
+    if (sigaction(SIGWINCH, NULL, &action) >= 0) {
+    	action.sa_flags |= SA_RESTART;
+    	sigaction(SIGWINCH, &action, NULL); /* Ignore errors */
+    }
+    return 0;
+}
+
+void
+get_file_dir_part(const char *word, char **dir_part, const char **file_part)
+{
+    const char *fp;
+
+    fp = strrchr(word, '/');
+    if (fp == NULL) {
+	*dir_part = xstrdup("");
+	*file_part = word;
+    } else {
+	for (; fp > word && fp[-1] == '/'; fp--);
+	*dir_part = xstrndup(word, fp - word + 1);
+	for (fp++; *fp == '/'; fp++);
+	*file_part = fp;
+    }
+}
+
+DCCompletionEntry *
+new_completion_entry(const char *input, const char *display)
+{
+    DCCompletionEntry *entry = xmalloc(sizeof(DCCompletionEntry));
+    entry->input = input == NULL ? NULL : xstrdup(input);
+    entry->display = display == NULL ? entry->input : xstrdup(display);
+    entry->input_char = ' ';
+    entry->input_format = "%s";
+    entry->input_single_format = NULL;
+    entry->input_single_full_format = NULL;
+    return entry;
+}
+
+void
+free_completion_entry(DCCompletionEntry *entry)
+{
+    if (entry->display != entry->input)
+        free(entry->display);
+    free(entry->input);
+    free(entry);
+}
+
+static char **
+attempted_completion(const char *text, int start, int end)
+{
+    char **matches = NULL;
+    char *word;
+    bool quoted;
+    PtrV *results = ptrv_new();
+
+    /* Determine if current word is quoted, and dequote. */
+    /*quoted = char_is_quoted(rl_line_buffer, start) && rl_line_buffer[start-1] == '"';
+    if (quoted)
+        start--;*/
+    quoted = rl_line_buffer[start] == '"';
+    word = dequote_words_full(rl_line_buffer+start, false, false, true, true, rl_line_buffer+end);
+
+    /* Do not attempt readline's default filename completion if our
+     * completors fails to return any results.
+     */
+    rl_attempted_completion_over = 1;
+    default_completion_selector(rl_line_buffer,start,end,word,results);
+
+    if (results->cur == 1) {
+        DCCompletionEntry *entry = results->buf[0];
+        char *match;
+
+        if (entry->input_single_full_format != NULL && strcmp(word, entry->input) == 0) {
+            match = xasprintf(entry->input_single_full_format, entry->input);
+        } else if (entry->input_single_format != NULL) {
+	    match = xasprintf(entry->input_single_format, entry->input);
+	} else {
+            match = xasprintf(entry->input_format, entry->input);
+        }
+	matches = xmalloc(sizeof(char *) * 2);
+        matches[0] = quote_word_full(match, quoted, true, ";", "#", false, true, true, true);
+        matches[1] = NULL;
+        rl_completion_append_character = entry->input_char;
+        free_completion_entry(entry);
+        free(match);
+    } else if (results->cur > 1) {
+        char *match = NULL;
+        int match_len = 0;
+        int c;
+
+    	matches = xmalloc(sizeof(char *) * (results->cur + 2));
+        for (c = 0; c < results->cur; c++) {
+            DCCompletionEntry *entry = results->buf[c];
+
+            /* Determine common leading characters for all matches. */
+            if (c == 0) {
+                match = xstrdup(entry->input);
+                match_len = strlen(match);
+            } else {
+                int d;
+                char c1, c2;
+                for (d = 0; (c1 = match[d]) && (c2 = entry->input[d]); d++) {
+	            if (c1 != c2)
+	    	        break;
+                }
+                match_len = MIN(match_len, d);
+            }
+
+	    matches[c+1] = quote_word_full(entry->display, false, true, "", "", false, true, true, false);
+	    free_completion_entry(entry);
+        }
+
+        match[match_len] = '\0';
+        rl_completion_append_character = ' ';
+	matches[0] =  quote_word_full(match, quoted, true, ";", "#", false, true, true, true);
+        matches[c+1] = NULL;
+        free(match);
+    }
+
+    ptrv_free(results);
+    free(word);
+    return matches;
+}
+
+/* This function is called via the warn_writer variable by warn() and
+ * die() in error.c to print messages on screen properly.
+ */
+static int
+screen_warn_writer(const char *format, va_list args)
+{
+    screen_writer(DC_DF_COMMON, format, args);
+    return 0;
+}
+
+static void
+flag_vputf(DCDisplayFlag flag, const char *format, va_list args)
+{
+    if (display_flags & flag) {
+        if (screen_state == SCREEN_SUSPENDED) {
+            ptrv_append(suspend_msgs, xvasprintf(format, args));
+        } else {
+            if (screen_state == SCREEN_RL_DISPLAYED) {
+                clear_rl();
+                screen_state = SCREEN_RL_CLEARED;
+            }
+            vprintf(format, args);
+        }
+    }
+    if (log_fh != NULL && log_flags & flag)
+        vfprintf(log_fh, format, args);
 }
 
 void
@@ -345,10 +307,7 @@ flag_putf(DCDisplayFlag flag, const char *format, ...)
     va_list args;
 
     va_start(args, format);
-    if (display_flags & flag)
-	screen_writer(format, args);
-    if (log_fh != NULL && log_flags & flag)
-	vfprintf(log_fh, format, args);
+    screen_writer(flag, format, args);
     va_end(args);
 }
 
@@ -359,7 +318,7 @@ static void
 user_input(char *line)
 {
     /* Readline has already made way for us. */
-    screen_state = SCREEN_NORMAL;
+    screen_state = SCREEN_RL_CLEARED;
 
     if (log_fh != NULL)
 	fprintf(log_fh, "> %s\n", line == NULL ? "(null)" : line);
@@ -370,14 +329,22 @@ user_input(char *line)
         running = false;
     } else if (line[0] != '\0') {
         add_history(line);
+	/* XXX: handle input differently
+	 * !shell_cmd
+	 * /cmd
+	 * depending on context: msg, say, cmd
+	 */
         command_execute(line);
     }
 
-    /* The following is necessary or readline will display
-     * a new prompt before we exit.
+    /* As soon as we exit this function, a new prompt will be displayed.
      */
-    if (!running) {
+    if (running) {
+        if (screen_state != SCREEN_SUSPENDED)
+            screen_state = SCREEN_RL_DISPLAYED;
+    } else {
         rl_callback_handler_remove();
+        FD_CLR(STDIN_FILENO, &read_fds);
         screen_state = SCREEN_NO_HANDLER;
     }
 }
@@ -392,26 +359,66 @@ user_input(char *line)
 void
 screen_erase_and_new_line(void)
 {
-    if (screen_state != SCREEN_NO_HANDLER) {
+    if (screen_state == SCREEN_RL_DISPLAYED) {
         rl_callback_handler_remove();
-        printf("\n");
+        putchar('\n');
         rl_callback_handler_install(screen_prompt, user_input);
     }
 }
 
+/* Suspend screen operations so that nothing in here
+ * uses standard in or standard out. This is necessary when
+ * running a system command in the terminal.
+ */
+void
+screen_suspend(void)
+{
+    if (screen_state == SCREEN_RL_DISPLAYED || screen_state == SCREEN_RL_CLEARED) {
+        rl_callback_handler_remove();
+	FD_CLR(STDIN_FILENO, &read_fds);
+        if (screen_state == SCREEN_RL_DISPLAYED)
+            putchar('\n');
+	suspend_msgs = ptrv_new();
+	screen_state = SCREEN_SUSPENDED;
+    }
+}
+
+/* Wake up screen after suspension.
+ */
+void
+screen_wakeup(bool print_newline_first)
+{
+    if (screen_state == SCREEN_SUSPENDED) {
+        int c;
+        if (print_newline_first)
+            putchar('\n');
+        for (c = 0; c < suspend_msgs->cur; c++)
+            fputs((char *) suspend_msgs->buf[c], stdout);
+	ptrv_foreach(suspend_msgs, free);
+	ptrv_free(suspend_msgs);
+        screen_state = SCREEN_NO_HANDLER;
+    }
+}
+
 /* Finish screen management. Usually called from main_finish.
  */
 void
 screen_finish(void)
 {
-    if (screen_state > SCREEN_NO_HANDLER) {
+    if (screen_state == SCREEN_SUSPENDED) {
+        int c;
+        for (c = 0; c < suspend_msgs->cur; c++)
+            fputs((char *) suspend_msgs->buf[c], stdout);
+	ptrv_foreach(suspend_msgs, free);
+	ptrv_free(suspend_msgs);
+    } else if (screen_state > SCREEN_NO_HANDLER) {
         rl_callback_handler_remove();
         if (screen_state == SCREEN_RL_DISPLAYED)
-    	    printf("\n");
-	screen_state = SCREEN_NO_HANDLER;
+            putchar('\n');
+	FD_CLR(STDIN_FILENO, &read_fds);
     }
 
-    if (screen_state > SCREEN_NO_HISTORY) {
+    if (screen_state >= SCREEN_NO_HANDLER) {
     	char *path;
 
 	/* Save history */
@@ -430,7 +437,8 @@ screen_finish(void)
     	rl_pre_input_hook = NULL;
 
 	warn_writer = default_warn_writer;
-        screen_state = SCREEN_NO_HISTORY;
+        screen_state = SCREEN_UNINITIALIZED;
+        free(screen_prompt);
 
 	set_log_file(NULL, false);
     }
@@ -442,17 +450,20 @@ screen_finish(void)
 void
 screen_prepare(void)
 {
-    if (screen_state <= SCREEN_NO_HISTORY) {
+    if (screen_state == SCREEN_SUSPENDED)
+        return;
+
+    if (screen_state == SCREEN_UNINITIALIZED) {
     	char *path;
 
     	screen_state = SCREEN_NO_HANDLER;
-	warn_writer = real_screen_vputf;
+	warn_writer = screen_warn_writer;
 	screen_prompt = xasprintf("%s> ", PACKAGE);
 
 	rl_readline_name = PACKAGE;
 	rl_attempted_completion_function = attempted_completion;
 	rl_completer_quote_characters = "\"";
-	rl_completer_word_break_characters = " \t\n\"";
+	rl_completer_word_break_characters = " \t\n\";";
 	/* rl_filename_quote_characters = " \t\n\\\"'>;|&()*?[]~!"; */
 	/* rl_filename_quoting_function = quote_argument; */
 	/* rl_filename_dequoting_function = dequote_argument_rl; */
@@ -466,8 +477,10 @@ screen_prepare(void)
     }
     if (screen_state == SCREEN_NO_HANDLER) {
         rl_callback_handler_install(screen_prompt, user_input);
+        FD_SET(STDIN_FILENO, &read_fds);
     }
     else if (screen_state == SCREEN_RL_CLEARED) {
+        rl_set_prompt(screen_prompt);
         rl_redisplay();
     }
     screen_state = SCREEN_RL_DISPLAYED;
@@ -501,31 +514,27 @@ set_screen_prompt(const char *prompt, ...)
     va_start(args, prompt);
     screen_prompt = xvasprintf(prompt, args);
     va_end(args);
-    rl_set_prompt(screen_prompt);
+    if (screen_state != SCREEN_SUSPENDED)
+        rl_set_prompt(screen_prompt);
 }
 
 /* Look up completion alternatives from a sorted list using binary search.
  */
-char *
-sorted_list_completion_generator(const char *base, int state, void *items,
+void
+sorted_list_completion_generator(const char *base, PtrV *results, void *items,
                                  size_t item_count, size_t item_size,
                                  size_t key_offset)
 {
-    static const void *item;
-    static const void *last_item;
+    const void *item;
+    const void *last_item;
 
-    if (state == 0) {
-	if (!bksearchrange(base, items, item_count, item_size,
-	                   key_offset, (comparison_fn_t) strleftcmp,
-	                   &item, &last_item))
-	    return NULL;
+    if (bksearchrange(base, items, item_count, item_size,
+	              key_offset, (comparison_fn_t) strleftcmp,
+	              &item, &last_item)) {
+        while (item <= last_item) {
+	    char *key = xstrdup(*(const char **) (((const char *) item) + key_offset));
+	    ptrv_append(results, new_completion_entry(key, NULL));
+	    item = (const void *) (((const char *) item) + item_size);
+        }
     }
-
-    if (item <= last_item) {
-	char *key = xstrdup(*(const char **) (((const char *) item) + key_offset));
-	item = (const void *) (((const char *) item) + item_size);
-	return key;
-    }
-
-    return NULL;
 }
