@@ -21,8 +21,9 @@
 #include <stdint.h>		/* Gnulib/POSIX/C99 */
 #include <sys/types.h>		/* ? */
 #include <sys/stat.h>		/* ? */
-#include <unistd.h>		/* POSIX */
+#include <unistd.h>		/* POSIX: confstr, ... */
 #include <fcntl.h>		/* ? */
+#include <string.h>		/* C89 */
 #include "xalloc.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
@@ -75,21 +76,39 @@ mkdirs_for_file(char *filename, bool deletedirs)
     return 0;
 }
 
-/* Can never return NULL */
+/* Concatenate two file names, adding slash character in between
+ * if necessary. An empty string for file name will be treated as
+ * the current directory, only invisible instead of `./'. If both
+ * P1 and P2 are empty strings, then the empty string will be
+ * returned. P1 and P2 must not be NULL. The returned value will
+ * be allocated on the heap and must be freed with free(). NULL
+ * will never be returned.
+ *
+ *	P1	P2	with slash	as is
+ * 	""	""	""		""
+ *	"foo"	""	"foo/"		"foo"
+ * 	"foo/"	""	"foo/"		"foo/"
+ *	""	"foo"	"foo/"		"foo"
+ *	""	"foo/"	"foo/		"foo/"
+ *	"foo"	"foo"	"foo/foo/"	"foo/foo"
+ *	"foo"	"foo/"	"foo/foo/"	"foo/foo"
+ *	"foo/"	"foo"	"foo/foo/"	"foo/foo/"
+ *	"foo/"	"foo/"	"foo/foo"	"foo/foo/"
+ */
+char *
+catfiles_with_trailing_slash(const char *p1, const char *p2)
+{
+    return xasprintf("%s%s%s%s",
+	    p1, p1[0] == '\0' || p1[strlen(p1)-1] == '/' ? "" : "/",
+	    p2, p2[0] == '\0' || p2[strlen(p2)-1] == '/' ? "" : "/");
+}
+
 char *
 catfiles(const char *p1, const char *p2)
 {
-    char *sep;
-    char *out;
-
-    if (*p2 == '\0')
-    	return xstrdup(p1);
-    if (*p1 == '\0')
-    	return xstrdup(p2);
-
-    sep = (p1[strlen(p1)-1] == '/' ? "" : "/");
-    out = xasprintf("%s%s%s", p1, sep, p2);
-    return out;
+    return xasprintf("%s%s%s",
+	    p1, p1[0] == '\0' || p1[strlen(p1)-1] == '/' ? "" : "/",
+	    p2);
 }
 
 /* Get environment variable with default value if it is not set.
@@ -103,25 +122,40 @@ getenv_default(const char *name, char *defvalue)
     return (value == NULL ? defvalue : value);
 }
 
-/* Set or remove the O_NONBLOCK flag on the specified file descriptor.
- * Can be moved into lib directory: Finalize name, expand interface to any
- * flag?
+/* Set or remove status flags on the specified file descriptor
+ * using the F_SETFL fcntl command. This function will not set
+ * call fcntl if the flags are already set (this is determined
+ * by issuing the F_GETFL fcntl command). SET determines if the
+ * flags MODFLAGS are to be added (true) or removed (false).
  */
-int
-fd_set_nonblock_flag(int fd, int value)
+bool
+fd_set_status_flags(int fd, bool set, int modflags)
 {
-    int flags;
+    int curflags;
+    int newflags;
 
-    flags = fcntl(fd, F_GETFL, 0);
-    if (flags < 0)
-    	return -1;
-
-    if (value != 0)
-    	flags |= flags;
+    curflags = fcntl(fd, F_GETFL, 0);
+    if (curflags < 0)
+    	return false;
+    if (set)
+    	newflags = curflags | modflags;
     else
-    	flags &= ~flags;
+    	newflags = curflags & ~modflags;
+    if (newflags == curflags)
+	return true;
 
-    return fcntl(fd, F_SETFL, flags);
+    return fcntl(fd, F_SETFL, newflags) != -1;
+}
+
+char *
+in_addr_str(struct in_addr addr)
+{
+    static char buffer[16];
+    unsigned char *bytes;
+    bytes = (unsigned char *) &addr;
+    snprintf(buffer, sizeof(buffer), "%d.%d.%d.%d",
+    	    bytes[0], bytes[1], bytes[2], bytes[3]);
+    return buffer;
 }
 
 char *
@@ -175,3 +209,44 @@ wordwrap(char *str, size_t len, size_t first_width, size_t other_width)
 
     return out;
 }
+
+/* for Gnulib? */
+#if 0
+char *
+xconfstr(const char *name)
+{
+    char *buf;
+    size_t len;
+
+    len = confstr(name, NULL, 0);
+    if (len != 0) {
+	buf = xmalloc(len);
+	if (confstr(name, buf, len) == 0)
+	    return buf;
+	free(buf);
+    }
+    return NULL;
+}
+#endif
+
+char *
+join_strings(char **strs, int count, char mid)
+{
+    int c;
+    size_t len = count;
+    char *out;
+    char *p;
+
+    for (c = 0; c < count; c++)
+        len += strlen(strs[c]);
+    p = out = xmalloc(len);
+    for (c = 0; c < count ;c++) {
+        p = stpcpy(p, strs[c]);
+        *p = mid;
+        p++;
+    }
+    p--;
+    *p = '\0';
+
+    return out;
+}
