@@ -41,32 +41,30 @@
 #include "common/quoting.h"
 #include "microdc.h"
 
-static void var_set_nick(DCVariable *var, const char *new_value);
-static void var_set_password(DCVariable *var, const char *new_value);
-static void var_set_description(DCVariable *var, const char *new_value);
-static void var_set_tag(DCVariable *var, const char *new_value);
-static void var_set_speed(DCVariable *var, const char *new_value);
-static void var_set_email(DCVariable *var, const char *new_value);
-static void var_set_share_dir(DCVariable *var, const char *new_value);
-static void var_set_download_dir(DCVariable *var, const char *new_value);
-static void var_set_listing_dir(DCVariable *var, const char *new_value);
-static void var_set_slots(DCVariable *var, const char *new_value);
+static void var_set_nick(DCVariable *var, int argc, char **argv);
+static void var_set_password(DCVariable *var, int argc, char **argv);
+static void var_set_description(DCVariable *var, int argc, char **argv);
+static void var_set_tag(DCVariable *var, int argc, char **argv);
+static void var_set_speed(DCVariable *var, int argc, char **argv);
+static void var_set_email(DCVariable *var, int argc, char **argv);
+static void var_set_share_dir(DCVariable *var, int argc, char **argv);
+static void var_set_download_dir(DCVariable *var, int argc, char **argv);
+static void var_set_listing_dir(DCVariable *var, int argc, char **argv);
+static void var_set_slots(DCVariable *var, int argc, char **argv);
 static char *var_get_slots(DCVariable *var);
 static char *var_get_string(DCVariable *var);
-static bool display_completion_generator(const char *base, int state, DCCompletionEntry *ce);
-static char *speed_completion_generator(const char *base, int state);
 static char *var_get_bool(DCVariable *var);
-static void var_set_active(DCVariable *var, const char *new_value);
+static void var_set_active(DCVariable *var, int argc, char **argv);
 static char *var_get_listen_addr(DCVariable *var);
-static void var_set_listen_addr(DCVariable *var, const char *new_value);
+static void var_set_listen_addr(DCVariable *var, int argc, char **argv);
 static char *var_get_listen_port(DCVariable *var);
-static void var_set_listen_port(DCVariable *var, const char *new_value);
+static void var_set_listen_port(DCVariable *var, int argc, char **argv);
 static char *var_get_display_flags(DCVariable *var);
-static void var_set_display_flags(DCVariable *var, const char *new_value);
-static void var_set_log_file(DCVariable *var, const char *new_value);
-static char *speed_completion_generator(const char *base, int state);
-static char *variable_completion_generator(const char *base, int state);
-static bool display_completion_generator(const char *base, int state, DCCompletionEntry *ce);
+static void var_set_display_flags(DCVariable *var, int argc, char **argv);
+static void var_set_log_file(DCVariable *var, int argc, char **argv);
+static void speed_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+static void variable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+static void display_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
 
 typedef struct {
     DCDisplayFlag flag;
@@ -104,112 +102,112 @@ static DCVariable variables[] = {
     {
       "active",
       var_get_bool, var_set_active, &is_active,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "Enable if listening for remote connections"
     },
     {
       "description",
       var_get_string, var_set_description, &my_description,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "This is the description which is visible to other users of the hub."
     },
     {
       "display",
       var_get_display_flags, var_set_display_flags, &display_flags,
-      DC_CPL_CUSTOM, display_completion_generator,
+      display_completion_generator,
       NULL,
       "Types of messages to display on screen"
     },
     {
       "downloaddir",
       var_get_string, var_set_download_dir, &download_dir,
-      DC_CPL_FILE, local_dir_completion_generator,
+      local_dir_completion_generator,
       NULL,
       "Directory which files are downloaded to"
     },
     {
       "email",
       var_get_string, var_set_email, &my_email,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "The e-mail visible to other users of the hub"
     },
     {
       "listenaddr",
       var_get_listen_addr, var_set_listen_addr, &force_listen_addr,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "Address to send to clients"
     },
     {
       "listenport",
       var_get_listen_port, var_set_listen_port, &listen_port,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "Port to listen on for connections"
     },
     {
       "listingdir",
-      var_get_string, var_set_listing_dir, &download_dir,
-      DC_CPL_FILE, local_dir_completion_generator,
+      var_get_string, var_set_listing_dir, &listing_dir,
+      local_dir_completion_generator,
       NULL,
       "Directory where file listings are kept"
     },
     {
       "log",
       var_get_display_flags, var_set_display_flags, &log_flags,
-      DC_CPL_CUSTOM, display_completion_generator,
+      display_completion_generator,
       NULL,
       "Types of messages to log (if logfile set)"
     },
     {
       "logfile",
       var_get_string, var_set_log_file, &log_filename,
-      DC_CPL_FILE, local_path_completion_generator,
+      local_path_completion_generator,
       NULL,
       "File to log screen messages to (will be appeneded)"
     },
     {
       "nick",
       var_get_string, var_set_nick, &my_nick,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "This is the desired (but not necessarily the current) nick name."
     },
     {
       "password",
       var_get_string, var_set_password, &my_password,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "The optional password to pass to the hub."
     },
     {
       "sharedir",
       var_get_string, var_set_share_dir, &share_dir,
-      DC_CPL_FILE, local_path_completion_generator,
+      local_path_completion_generator,
       NULL,
       "Directory containing files to share (or a single file to share)"
     },
     {
       "slots",
       var_get_slots, var_set_slots, &my_ul_slots,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "Number of open upload slots"
     },
     {
       "speed",
       var_get_string, var_set_speed, &my_speed,
-      DC_CPL_SIMPLE, speed_completion_generator,
+      speed_completion_generator,
       NULL,
       "The speed visible to other users of the hub"
     },
     {
       "tag",
       var_get_string, var_set_tag, &my_tag,
-      DC_CPL_NONE, NULL,
+      NULL,
       NULL,
       "The user agent tag the hub uses to detect features"
     },
@@ -226,124 +224,154 @@ find_variable(const char *name)
 }
 
 static void
-var_set_nick(DCVariable *var, const char *new_value)
+var_set_nick(DCVariable *var, int argc, char **argv)
 {
-    if (*new_value == '\0') {
-        warn(_("Nick cannot be empty.\n"));
-    } else if (strnlen(new_value, 36) >= 36) {
-        warn(_("Nick is too long - max length is 35 characters.\n"));
-    } else if (strpbrk(new_value, "$| ") != NULL) {
-        warn(_("Nick may not contain `$', `|' or space characters.\n"));
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+    } else if (argv[1][0] == '\0') {
+	warn(_("Nick cannot be empty.\n"));
+    } else if (strnlen(argv[1], 36) >= 36) {
+	warn(_("Nick is too long - max length is 35 characters.\n"));
+    } else if (strpbrk(argv[1], "$| ") != NULL) {
+	warn(_("Nick may not contain `$', `|' or space characters.\n"));
     } else {
-        free(my_nick);
-        my_nick = xstrdup(new_value);
+	free(my_nick);
+	my_nick = xstrdup(argv[1]);
     }
 }
 
 static void
-var_set_description(DCVariable *var, const char *new_value)
+var_set_description(DCVariable *var, int argc, char **argv)
 {
+    char *new_value = join_strings(argv+1, argc-1, ' ');
+
     if (strpbrk(new_value, "$|") != NULL) {
         warn(_("Description may not contain `$' or `|' characters.\n"));
+	free(new_value);
     } else if (strnlen(new_value, 36) >= 36) {
         warn(_("Description is too long - max length is 35 characters.\n"));
+	free(new_value);
     } else {
 	free(my_description);
-	my_description = xstrdup(new_value);
+	my_description = new_value;
     }
+    
 }
 
 static void
-var_set_email(DCVariable *var, const char *new_value)
+var_set_email(DCVariable *var, int argc, char **argv)
 {
-    if (strpbrk(new_value, "$|") != NULL) {
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+    } else if (strpbrk(argv[1], "$|") != NULL) {
         warn(_("E-mail may not contain `$' or `|' characters.\n"));
-    } else if (strnlen(new_value, 36) >= 36) {
+    } else if (strnlen(argv[1], 36) >= 36) {
         warn(_("E-mail is too long - max length is 35 characters.\n"));
     } else {
         free(my_email);
-        my_email = xstrdup(new_value);
+        my_email = xstrdup(argv[1]);
     }
 }
 
 static void
-var_set_tag(DCVariable *var, const char *new_value)
+var_set_tag(DCVariable *var, int argc, char **argv)
 {
+    char *new_value = join_strings(argv+1, argc-1, ' ');
+
     if (strpbrk(new_value, "$|") != NULL) {
-        warn(_("Tag may not contain `$' or `|' characters."));
+        warn(_("Tag may not contain `$' or `|' characters.\n"));
+	free(new_value);
     } else {
         free(my_tag);
-        my_tag = xstrdup(new_value);
+        my_tag = new_value;
     }
 }
 
 static void
-var_set_speed(DCVariable *var, const char *new_value)
+var_set_speed(DCVariable *var, int argc, char **argv)
 {
-    if (strpbrk(new_value, "$|") != NULL) {
-        warn(_("Speed may not contain `$' or `|' characters."));
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+    } else if (strpbrk(argv[1], "$|") != NULL) {
+        warn(_("Speed may not contain `$' or `|' characters.\n"));
     } else {
         free(my_speed);
-        my_speed = xstrdup(new_value);
+        my_speed = xstrdup(argv[1]);
     }
 }
 
 static void
-var_set_share_dir(DCVariable *var, const char *new_value)
+var_set_share_dir(DCVariable *var, int argc, char **argv)
 {
     struct stat st;
 
-    if (stat(new_value, &st) < 0) {
-    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(new_value), errstr);
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (stat(argv[1], &st) < 0) {
+    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(argv[1]), errstr);
 	return;
     }
     if (!S_ISDIR(st.st_mode)) {
-    	screen_putf(_("%s: Not a directory\n"), quotearg(new_value));
+    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
 	return;
     }
-    set_share_dir(new_value); /* Ignore errors */
+    set_share_dir(argv[1]); /* Ignore errors */
 }
 
 static void
-var_set_download_dir(DCVariable *var, const char *new_value)
+var_set_download_dir(DCVariable *var, int argc, char **argv)
 {
     struct stat st;
 
-    if (stat(new_value, &st) < 0) {
-    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(new_value), errstr);
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (stat(argv[1], &st) < 0) {
+    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(argv[1]), errstr);
 	return;
     }
     if (!S_ISDIR(st.st_mode)) {
-    	screen_putf(_("%s: Not a directory\n"), quotearg(new_value));
+    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
 	return;
     }
     free(download_dir);
-    download_dir = xstrdup(new_value);
+    download_dir = xstrdup(argv[1]);
 }
 
 static void
-var_set_listing_dir(DCVariable *var, const char *new_value)
+var_set_listing_dir(DCVariable *var, int argc, char **argv)
 {
     struct stat st;
 
-    if (stat(new_value, &st) < 0) {
-    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(new_value), errstr);
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (stat(argv[1], &st) < 0) {
+    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(argv[1]), errstr);
 	return;
     }
     if (!S_ISDIR(st.st_mode)) {
-    	screen_putf(_("%s: Not a directory\n"), quotearg(new_value));
+    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
 	return;
     }
     free(listing_dir);
-    listing_dir = xstrdup(new_value);
+    listing_dir = xstrdup(argv[1]);
     set_share_dir(share_dir);
 }
 
 static void
-var_set_slots(DCVariable *var, const char *new_value)
+var_set_slots(DCVariable *var, int argc, char **argv)
 {
-    if (!parse_uint32(new_value, &my_ul_slots))
-    	screen_putf(_("Invalid slot number `%s'\n"), quotearg(new_value));
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (!parse_uint32(argv[1], &my_ul_slots))
+    	screen_putf(_("Invalid slot number `%s'\n"), quotearg(argv[1]));
 }
 
 static char *
@@ -373,10 +401,15 @@ parse_bool(const char *s, bool *value)
 }
 
 static void
-var_set_active(DCVariable *var, const char *new_value)
+var_set_active(DCVariable *var, int argc, char **argv)
 {
     bool state;
-    if (!parse_bool(new_value, &state)) {
+    
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (!parse_bool(argv[1], &state)) {
 	screen_putf(_("Specify active as `0', `no', `off', `1', `yes', or `on'.\n"));
 	return;/*false*/
     }
@@ -389,27 +422,31 @@ var_get_listen_addr(DCVariable *var)
 {
     if (force_listen_addr.s_addr == INADDR_NONE)
 	return NULL;
-    return xstrdup(sockaddr_in_str(&local_addr));
+    return xstrdup(in_addr_str(force_listen_addr));
 }
 
 static void
-var_set_listen_addr(DCVariable *var, const char *new_value)
+var_set_listen_addr(DCVariable *var, int argc, char **argv)
 {
     struct sockaddr_in addr;
 
-    if (*new_value == '\0') {
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (argv[1][0] == '\0') {
 	force_listen_addr.s_addr = INADDR_NONE;
 	screen_putf(_("Removing listening address.\n"));
 	return;
     }
 
-    if (!inet_aton(new_value, &addr.sin_addr)) {
+    if (!inet_aton(argv[1], &addr.sin_addr)) {
 	struct hostent *he;
 
-	screen_putf(_("Looking up IP address for %s\n"), quotearg(new_value));
-	he = gethostbyname(new_value);
+	screen_putf(_("Looking up IP address for %s\n"), quotearg(argv[1]));
+	he = gethostbyname(argv[1]);
 	if (he == NULL) {
-	    screen_putf(_("%s: Cannot look up address - %s\n"), quotearg(new_value), hstrerror(h_errno));
+	    screen_putf(_("%s: Cannot look up address - %s\n"), quotearg(argv[1]), hstrerror(h_errno));
 	    return;
 	}
 
@@ -453,16 +490,15 @@ find_display_flag_value(const char *name)
 }
 
 static void
-var_set_display_flags(DCVariable *var, const char *new_value)
+var_set_display_flags(DCVariable *var, int argc, char **argv)
 {
     uint32_t add_values = 0;
     uint32_t set_values = 0;
     uint32_t del_values = 0;
     uint32_t c;
-    char *arg;
 
-    /* XXX: better new_values[n] - but how!? */
-    for (c = 0; (arg = get_word_dequoted(new_value, c)) != NULL; c++) {
+    for (c = 1; c < argc; c++) {
+	char *arg = argv[c];
 	uint32_t *values;
 	uint32_t value;
 
@@ -496,6 +532,7 @@ var_set_display_flags(DCVariable *var, const char *new_value)
 	*((uint32_t *) var->value) = set_values;
     *((uint32_t *) var->value) |= add_values;
     *((uint32_t *) var->value) &= ~del_values;
+    *((uint32_t *) var->value) |= DC_DF_COMMON;	/* XXX: must always be available, but for logging? */
 }
 
 static char *
@@ -507,14 +544,18 @@ var_get_listen_port(DCVariable *var)
 }
 
 static void
-var_set_listen_port(DCVariable *var, const char *new_value)
+var_set_listen_port(DCVariable *var, int argc, char **argv)
 {
     uint16_t port;
-    if (*new_value == '\0') {
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (argv[1][0] == '\0') {
 	port = 0;
     } else {
-	if (!parse_uint16(new_value, &port)) {
-	    screen_putf(_("Invalid value `%s' for port number.\n"), quotearg(new_value));
+	if (!parse_uint16(argv[1], &port)) {
+	    screen_putf(_("Invalid value `%s' for port number.\n"), quotearg(argv[1]));
 	    return;
 	}
     }
@@ -523,9 +564,13 @@ var_set_listen_port(DCVariable *var, const char *new_value)
 }
 
 static void
-var_set_log_file(DCVariable *var, const char *new_value)
+var_set_log_file(DCVariable *var, int argc, char **argv)
 {
-    set_log_file(new_value, true);
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    set_log_file(argv[1], true);
 }
 
 static char *
@@ -543,17 +588,21 @@ var_get_bool(DCVariable *var)
 }
 
 static void
-var_set_password(DCVariable *var, const char *new_value)
+var_set_password(DCVariable *var, int argc, char **argv)
 {
-    if (*new_value == '\0') {
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (argv[1][0] == '\0') {
 	screen_putf(_("Removing current password.\n"));
 	free(my_password);
 	my_password = NULL;
-    } else if (strchr(new_value, '|') != NULL) {
-    	screen_putf(_("Password may not contain `|' characters."));
+    } else if (strchr(argv[1], '|') != NULL) {
+    	warn(_("Password may not contain `|' characters.\n"));
     } else {
 	free(my_password);
-	my_password = xstrdup(new_value);
+	my_password = xstrdup(argv[1]);
     }
 }
 
@@ -561,74 +610,66 @@ var_set_password(DCVariable *var, const char *new_value)
  * binsearch for the first match, then stopping immediately when a
  * non-match is found.
  */
-static char *
-variable_completion_generator(const char *base, int state)
+void
+variable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    return sorted_list_completion_generator(base, state, variables, variables_count, sizeof(DCVariable), offsetof(DCVariable, name));
+    sorted_list_completion_generator(word,results,variables,variables_count,sizeof(DCVariable),offsetof(DCVariable, name));
 }
 
-static char *
-speed_completion_generator(const char *base, int state)
+static void
+speed_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static int completion_pos;
     int c;
 
     /* There's not enough speed types to justify a binary search... */
-    if (state == 0)
-	completion_pos = 0;
-
-    for (c = completion_pos; c < speeds_count; c++) {
-	if (strleftcmp(base, speeds[c]) == 0) {
-	    completion_pos = c+1;
-	    return xstrdup(speeds[c]); /* no need for quoting */
-	}
+    for (c = 0; c < speeds_count; c++) {
+	if (strleftcmp(word, speeds[c]) == 0)
+	    ptrv_append(results, new_completion_entry(speeds[c], NULL));
     }
-
-    return NULL;
 }
 
-static bool
-display_completion_generator(const char *base, int state, DCCompletionEntry *ce)
+static void
+display_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
-    static int completion_pos;
     char base_char;
     uint32_t c;
+    uint32_t flags;
+    char *cmd;
 
-    base_char = (base[0] == '+' || base[0] == '-' ? *base++ : '\0');
-    if (state == 0)
-        completion_pos = 0;
+    cmd = get_word_dequoted(line, 1);
+    if (strcmp(cmd, "log") == 0) { /* this is safe even when using aliases */
+	flags = log_flags;
+    } else {
+	flags = display_flags;
+    }
+    free(cmd);
 
+    base_char = (word[0] == '+' || word[0] == '-' ? *word++ : '\0');
     /* XXX: complete `all', `default' */
+    for (c = 0; c < display_flag_count; c++) {
+        if (strleftcmp(word, display_flag_details[c].name) == 0) {
+            DCCompletionEntry *entry;
+        
+            if (base_char == '+' && (flags & display_flag_details[c].flag) != 0)
+                continue;
+            if (base_char == '-' && (flags & display_flag_details[c].flag) == 0)
+                continue;
 
-    for (c = completion_pos; c < display_flag_count; c++) {
-        if (strleftcmp(base, display_flag_details[c].name) == 0) {
-            if (base_char == '+' && (display_flags & display_flag_details[c].flag) != 0)
-                continue;
-            if (base_char == '-' && (display_flags & display_flag_details[c].flag) == 0)
-                continue;
-            if (base_char == '+')
-                ce->input_format = "+%s";
-            else if (base_char == '-')
-                ce->input_format = "-%s";
-            ce->str = xstrdup(display_flag_details[c].name); /* no need for quoting */
-	    completion_pos = c+1;
-            return true;
+            entry = new_completion_entry(NULL, NULL);
+            entry->input = xasprintf("%c%s", base_char, display_flag_details[c].name);
+            entry->display = xstrdup(display_flag_details[c].name);
+	    ptrv_append(results, entry);
         }
     }
-
-    return false;
 }
 
 void
-cmd_set(char *cmd, char *args)
+cmd_set(int argc, char **argv)
 {
-    char *name;
-    char *value;
     uint32_t c;
     DCVariable *var;
-    
-    name = (args == NULL ? NULL : get_word_dequoted(args, 0));
-    if (name == NULL) {
+
+    if (argc == 1) {
 	int max_len = 0;
 	int cols;
 	char *fmt;
@@ -648,14 +689,14 @@ cmd_set(char *cmd, char *args)
 	return;
     }
 
-    var = find_variable(name);
+    var = find_variable(argv[1]);
     if (var == NULL) {
-	warn("No variable by the name `%s'.\n", quotearg(name));
+	warn("No variable by the name `%s'.\n", quotearg(argv[1]));
 	return;
     }
 
-    value = get_words_dequoted(args, 1, SIZE_MAX);
-    if (value == NULL) {
+    if (argc <= 2) {
+	char *value;
 	value = var->getter(var);
 	if (value == NULL) {
 	    screen_putf(_("No value is set for `%s'.\n"), var->name);
@@ -665,32 +706,23 @@ cmd_set(char *cmd, char *args)
 	return;
     }
 
-    var->setter(var, value);
+    var->setter(var, argc-1, argv+1);
 }
 
-void *
-set_completion_selector(const char *buffer, int start, int end, DCCompletionFunctionType *type)
+void
+set_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
 {
     char *name;
-    int c;
     DCVariable *var;
 
-    /* XXX: return type instead...? */
-
-    c = get_word_index(buffer, end);
-    if (c <= 1) { /* c > 0 is assured by get_completor. */
-	*type = DC_CPL_SIMPLE;
-	return variable_completion_generator;
+    /* get_word_index(..) > 0 is assured by get_completor. */
+    if (get_word_index(line, we) <= 1) { 
+	variable_completion_generator(line,ws,we,word,results);
+    } else {
+	name = get_word_dequoted(line, 1);
+	var = find_variable(name);
+	if (var != NULL && var->completor != NULL)
+	    var->completor(line,ws,we,word,results);
+	free(name);
     }
-
-    name = get_word_dequoted(buffer, 1);
-    /* name != NULL is assured because of get_word_index > 1 */
-    var = find_variable(name);
-    if (var == NULL) {
-	*type = DC_CPL_NONE;
-	return NULL;
-    }
-
-    *type = var->completor_type;
-    return var->completor;
 }
